/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998-2000  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * linkfd.c,v 1.4.2.11 2002/04/25 09:19:50 bergolth Exp
 */

#include "config.h"
 
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <strings.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <syslog.h>
#include <time.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SCHED_H
#include <sched.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include "vtun.h"
#include "linkfd.h"
#include "lib.h"
#include "driver.h"

char * get_local_addr( struct vtun_host *host, int i);
void Idle( int i, char *log);
void Send_Close_All( int i);
/* Host we are working with. 
 * Used by signal handlers that's why it is global. 
 */
struct vtun_host *lfd_host;

struct lfd_mod *lfd_mod_head = NULL, *lfd_mod_tail = NULL;

/* Modules functions*/

/* Add module to the end of modules list */
void lfd_add_mod(struct lfd_mod *mod)
{
     if( !lfd_mod_head ){
        lfd_mod_head = lfd_mod_tail = mod;
	mod->next = mod->prev = NULL;
     } else {
        lfd_mod_tail->next = mod;
        mod->prev = lfd_mod_tail;
        mod->next = NULL;
        lfd_mod_tail = mod;
     }
}

/*  Initialize and allocate each module */
int lfd_alloc_mod(struct vtun_host *host)
{
     struct lfd_mod *mod = lfd_mod_head;

     while( mod ){
        if( mod->alloc && (mod->alloc)(host) )
	   return 1; 
	mod = mod->next;
     } 

     return 0;
}

/* Free all modules */
int lfd_free_mod(void)
{
     struct lfd_mod *mod = lfd_mod_head;

     while( mod ){
        if( mod->free && (mod->free)() )
	   return 1;
	mod = mod->next;
     } 
     lfd_mod_head = lfd_mod_tail = NULL;
     return 0;
}

 /* Run modules down (from head to tail) */
inline int lfd_run_down(int len, char *in, char **out)
{
     register struct lfd_mod *mod;
     
     *out = in;
     for(mod = lfd_mod_head; mod && len > 0; mod = mod->next )
        if( mod->encode ){
           len = (mod->encode)(len, in, out);
           in = *out;
        }
     return len;
}

/* Run modules up (from tail to head) */
inline int lfd_run_up(int len, char *in, char **out)
{
     register struct lfd_mod *mod;
     
     *out = in;
     for(mod = lfd_mod_tail; mod && len > 0; mod = mod->prev )
        if( mod->decode ){
	   len = (mod->decode)(len, in, out);
           in = *out;
	}
     return len;
}

/* Check if modules are accepting the data(down) */
inline int lfd_check_down(void)
{
     register struct lfd_mod *mod;
     int err = 1;
 
     for(mod = lfd_mod_head; mod && err > 0; mod = mod->next )
        if( mod->avail_encode )
           err = (mod->avail_encode)();
     return err;
}

/* Check if modules are accepting the data(up) */
inline int lfd_check_up(void)
{
     register struct lfd_mod *mod;
     int err = 1;

     for(mod = lfd_mod_tail; mod && err > 0; mod = mod->prev)
        if( mod->avail_decode )
           err = (mod->avail_decode)();

     return err;
}
		
/********** Linker *************/
/* Termination flag */
static volatile sig_atomic_t linker_term;

static void sig_term(int sig)
{
     vtun_syslog(LOG_INFO, "Closing connection");
     io_cancel();
     linker_term = VTUN_SIG_TERM;
}

static void sig_hup(int sig)
{
     vtun_syslog(LOG_INFO, "Reestablishing connection");
     io_cancel();
     linker_term = VTUN_SIG_HUP;
}

/* Statistic dump */
int live_host_num; 
char live_host_str[10]; 
unsigned long In, Out;
time_t tm_IO;
int jitter[5*2+1];
void Set_IO()
{
	time_t tm=time(NULL);
	time_t gap= tm - tm_IO;
	time_t gap2= tm - lfd_host->stat.tm_old;
	if( gap  < 1) gap=1;
	if( gap2 < 1) gap=1;
	if( gap  < 6) return; /* low gap */
        set_title("%-6s %-5s %-4s %4luk %4luk (%4luk %4luk in %lu:%02lu) (%lu/%lu)(%d/%d)", 
		lfd_host->host, lfd_host->sopt->dev, 
		live_host_str,
		In*8/gap/1000, Out*8/gap/1000,
		lfd_host->stat.byte_in*8/gap2/1000, lfd_host->stat.byte_out*8/gap2/1000, gap2/60, gap2%60,
		lfd_host->stat.packet_loss, lfd_host->stat.packet_count,
		live_host_num, vtun.svr_num 
		);

	/*reset counter*/
	In=Out=0;
	tm_IO=tm;
	
}
void Set_title()
{
     /*
     time_t gap= time(NULL) - lfd_host->stat.tm_old;
     if( gap < 1) gap=1;
     set_title("%s %s (%d/%d)%s %luk %luk in %lum%lus", lfd_host->host, lfd_host->sopt->dev, live_host_num, vtun.svr_num, live_host_str,
		lfd_host->stat.byte_in*8/gap/1000, lfd_host->stat.byte_out*8/gap/1000, gap/60, gap%60	);
     */
     Set_IO();
}
void log_stat( char * stat_in)
{
     static time_t tm;
     static char stm[22];
     time_t gap;
     int i;
  
     tm = time(NULL);
     gap= tm - lfd_host->stat.tm_old;
     if( gap < 10) return; /* low gap then return */ 

     strftime(stm, sizeof(stm)-1, "%Y/%m/%d %H:%M:%S", localtime(&tm));
     /* 20040401 11:11:11 UUUU gap in_loss out_loss in_count out_count in_speed out_speed */
     fprintf(lfd_host->stat.file,"%s %s %3lu %4lu %5lu %3lu %6lu %s\n", stm, live_host_str,
		gap,
		lfd_host->stat.byte_in*8/gap/1000, lfd_host->stat.byte_out*8/gap/1000, 
		lfd_host->stat.packet_loss, lfd_host->stat.packet_count,
		(stat_in? stat_in : "0    0  END 0")
		);

     /* reset to zero */
     lfd_host->stat.byte_in = lfd_host->stat.byte_out = 0;
     lfd_host->stat.comp_in = lfd_host->stat.comp_out = 0; 
     lfd_host->stat.packet_loss = lfd_host->stat.packet_count = 0; 
     lfd_host->stat.tm_old= tm;
     for( i=1; i<=vtun.svr_num; i++){
	jitter[i]=0;
     }
}
int idle[5*2+1];
int fd[5*2+1], order[5+1];
char *buf[5*2+1], *out;

void send_log()
{
     static time_t tm;
     time_t gap;
     int i;
  
     tm = time(NULL);
     gap= tm - lfd_host->stat.tm_old;
     if( gap < 10) return; /* low gap then return */ 

     sprintf(buf[0],"%3lu %6lu %s %d", 
		lfd_host->stat.packet_loss, lfd_host->stat.packet_count,
		live_host_str, gap
		);

     /*
     vtun_syslog(LOG_INFO,"from_server stat_req");
     */
     for( i=1; i<=vtun.svr_num; i++){
	jitter[i]=0;
	if( idle[i] != -1){
     	    if( udp_write2(fd[i], buf[0], VTUN_STAT_REP+strlen(buf[0])+1, 0) < 0 ) Idle(i,"stat_rep");
	}
     }

     /* reset to zero */
     lfd_host->stat.byte_in = lfd_host->stat.byte_out = 0;
     lfd_host->stat.comp_in = lfd_host->stat.comp_out = 0; 
     lfd_host->stat.packet_loss = lfd_host->stat.packet_count = 0; 
     lfd_host->stat.tm_old= tm;
}
     
void sig_alarm(int sig)
{
     int k;
     /* send request */
     for( k=1; k<=vtun.svr_num; k++){
	 if( idle[k] != -1){
              /*
              vtun_syslog(LOG_INFO,"STAT_REQ %d",k);
	      */
              if( udp_write2(fd[k], buf[k], VTUN_STAT_REQ, 0) < 0 ) Idle(k,"send stat req error");
	 }
     }
     alarm(VTUN_STAT_IVAL);
}    

#define VTUN_DIR "/etun"
void do_test( int do_num)
{
}

static void sig_usr1(int sig)
{
     /* Start Test module using sky00.cmd*/
	/*
	FILE *fp;
	char file[50];
	int do_num=0,k;
	sprintf(file,"%s/%.20s.do", VTUN_DIR, lfd_host->host);
	fp= fopen(file,"r");
	if( fp == NULL){
              vtun_syslog(LOG_INFO,"TEST_REQ can't read do file %s",file);
	      return;
	}else{
	      do_num= atoi(fgets(file,40 ,fp));
	      if( do_num <1 || do_num > vtun.svr_num){
                      vtun_syslog(LOG_INFO,"TEST_REQ Bad do line num %s",file);
		      return;
	      }
              vtun_syslog(LOG_INFO,"TEST_REQ Try line num %d", do_num);
              for( k=1; k<=vtun.svr_num; k++){
	         if( idle[k] != -1){
                   vtun_syslog(LOG_INFO,"TEST_REQ to line %d for test line %d",k, do_num);
                  if( udp_write2(fd[k], buf[k], VTUN_TEST_REQ, do_num) < 0 ) Idle(k,"send test req error");
	         }
              }
	      do_test(do_num);
	}
	*/
}

int lfd_linker(void)
{
     int fd1 = lfd_host->tcp_fd;
     int fd2 = lfd_host->loc_fd; 
     register int len, fl;
     struct timeval tv;
     char *buf, *out;
     fd_set fdset;
     int maxfd, idle = 0;
     time_t fd1_last=time(NULL);

     if( !(buf = lfd_alloc(VTUN_FRAME_SIZE + VTUN_FRAME_OVERHEAD)) ){
	vtun_syslog(LOG_ERR,"Can't allocate buffer for the linker"); 
        return 0; 
     }

     proto_write(fd1, buf, VTUN_ECHO_REQ);

     maxfd = (fd1 > fd2 ? fd1 : fd2) + 1;

     linker_term = 0;
     while( !linker_term ){
	errno = 0;

        /* Wait for data */
        FD_ZERO(&fdset);
	FD_SET(fd1, &fdset);
	FD_SET(fd2, &fdset);

 	tv.tv_sec  = lfd_host->ka_interval;
	tv.tv_usec = 0;

	if( (len = select(maxfd, &fdset, NULL, NULL, &tv)) < 0 ){
	   if( errno != EAGAIN && errno != EINTR )
	      break;
	   else
	      continue;
	} 
	
	if( !FD_ISSET(fd1, &fdset) && ((time(NULL)-fd1_last) > lfd_host->ka_interval) ){
	fd1_last= time(NULL);
/*
        } 
	if( !len ){
*/
	   /* We are idle, lets check connection */
	   if( lfd_host->flags & VTUN_KEEP_ALIVE ){
	      if( ++idle > lfd_host->ka_failure ){
	         vtun_syslog(LOG_INFO,"Session %s network timeout", lfd_host->host);
		 break;	
	      }
	      /* Send ECHO request */
	      if( proto_write(fd1, buf, VTUN_ECHO_REQ) < 0 )
		 break;
	   }
	   continue;
	}	   

	/* Read frames from network(fd1), decode and pass them to 
         * the local device (fd2) */
	if( FD_ISSET(fd1, &fdset) && lfd_check_up() ){
	   fd1_last= time(NULL);
	   idle = 0; 
	   if( (len=proto_read(fd1, buf)) <= 0 )
	      break;

	   /* Handle frame flags */
	   fl = len & ~VTUN_FSIZE_MASK;
           len = len & VTUN_FSIZE_MASK;
	   if( fl ){
	      if( fl==VTUN_BAD_FRAME ){
		 vtun_syslog(LOG_ERR, "Received bad frame");
		 continue;
	      }
	      if( fl==VTUN_ECHO_REQ ){
		 /* Send ECHO reply */
	 	 if( proto_write(fd1, buf, VTUN_ECHO_REP) < 0 )
		    break;
		 continue;
	      }
   	      if( fl==VTUN_ECHO_REP ){
		 /* Just ignore ECHO reply */
		 continue;
	      }
	      if( fl==VTUN_CONN_CLOSE ){
	         vtun_syslog(LOG_INFO,"Connection closed by other side");
		 break;
	      }
	   }   

	   lfd_host->stat.comp_in += len; 
	   if( (len=lfd_run_up(len,buf,&out)) == -1 )
	      break;	
	   if( len && dev_write(fd2,out,len) < 0 ){
              if( errno != EAGAIN && errno != EINTR )
                 break;
              else
                 continue;
           }
	   lfd_host->stat.byte_in += len; 
	}

	/* Read data from the local device(fd2), encode and pass it to 
         * the network (fd1) */
	if( FD_ISSET(fd2, &fdset) && lfd_check_down() ){
	   if( (len = dev_read(fd2, buf, VTUN_FRAME_SIZE)) < 0 ){
	      if( errno != EAGAIN && errno != EINTR )
	         break;
	      else
		 continue;
	   }
	   if( !len ) break;
	
	   lfd_host->stat.byte_out += len; 
	   if( (len=lfd_run_down(len,buf,&out)) == -1 )
	      break;
	   if( len && proto_write(fd1, out, len) < 0 )
	      break;
	   lfd_host->stat.comp_out += len; 
	}
     }
     if( !linker_term && errno )
	vtun_syslog(LOG_INFO,"%s (%d)", strerror(errno), errno);

     if (linker_term == VTUN_SIG_TERM) {
       lfd_host->persist = 0;
     }

     /* Notify other end about our close */
     proto_write(fd1, buf, VTUN_CONN_CLOSE);
     lfd_free(buf);

     return 0;
}

/* Link remote and local file descriptors */ 
int linkfd(struct vtun_host *host)
{
     struct sigaction sa, sa_oldterm, sa_oldint, sa_oldhup;
     int old_prio;

     lfd_host = host;
 
     old_prio=getpriority(PRIO_PROCESS,0);
     setpriority(PRIO_PROCESS,0,LINKFD_PRIO);

     /* Build modules stack */
     if(host->flags & VTUN_ZLIB)
	lfd_add_mod(&lfd_zlib);

     if(host->flags & VTUN_LZO)
	lfd_add_mod(&lfd_lzo);

     if(host->flags & VTUN_ENCRYPT)
	lfd_add_mod(&lfd_encrypt);

     if(host->flags & VTUN_SHAPE)
	lfd_add_mod(&lfd_shaper);

     if(lfd_alloc_mod(host))
	return 0;

     memset(&sa, 0, sizeof(sa));
     sa.sa_handler=sig_term;
     sigaction(SIGTERM,&sa,&sa_oldterm);
     sigaction(SIGINT,&sa,&sa_oldint);
     sa.sa_handler=sig_hup;
     sigaction(SIGHUP,&sa,&sa_oldhup);

     /* It will use for test module*/
     sa.sa_handler=sig_usr1;
     sigaction(SIGUSR1,&sa,NULL);
     /* Initialize statstic dumps when it is server*/
     if( (host->flags & VTUN_STAT) && !vtun.is_client ){
	char file[40];

        sa.sa_handler=sig_alarm;
        sigaction(SIGALRM,&sa,NULL);

	sprintf(file,"%s/%.20s", VTUN_STAT_DIR, host->host);
	if( (host->stat.file=fopen(file, "a")) ){
	   setvbuf(host->stat.file, NULL, _IOLBF, 0);
	   alarm(VTUN_STAT_IVAL);
	} else
	   vtun_syslog(LOG_ERR, "Can't open stats file %s", file);
     }

     io_init();

     if(vtun.svr_num > 0)
       lfd_linker_udp();
     else
       lfd_linker();

     if( (host->flags & VTUN_STAT) && !vtun.is_client ){
        alarm(0);
	if (host->stat.file)
	  fclose(host->stat.file);
     }

     lfd_free_mod();
     
     sigaction(SIGTERM,&sa_oldterm,NULL);
     sigaction(SIGINT,&sa_oldint,NULL);
     sigaction(SIGHUP,&sa_oldhup,NULL);

     setpriority(PRIO_PROCESS,0,old_prio);

     return linker_term;
}

static unsigned long inline tv2ms(struct timeval tv)
{
     register unsigned long ms = (tv.tv_sec * 1000)+(tv.tv_usec / 1000);
     return  ms ? ms : 1;
}

int send_close_try_count;
int send_close[5*2+1];
void Send_Close_All( int i)
{
    int j;
    for( j=1; j<=vtun.svr_num; j++){
      if( idle[j] != -1){
       vtun_syslog(LOG_INFO,"%s send close line %d using %d", lfd_host->host, i, j); 
       /* try close other side */
       udp_write2(fd[j], buf[j], VTUN_CONN_CLOSE, i);
      }
    }
}

time_t fd_idle;
time_t fd_idle_last;
void Idle( int i, char *log)
{
  if( log ) vtun_syslog(LOG_INFO,"%s network down p:%d (%s:%s) by:%s ", 
		 lfd_host->host, i, lfd_host->sopt[i-1].laddr, lfd_host->sopt[i-1].raddr, log ); 
  live_host_num--;
  if( live_host_num == 1) lfd_host->ka_failure= 10; else lfd_host->ka_failure=2;
  live_host_str[i-1]='D';
  Set_title();
  idle[i]=-1;
  fd_idle= time(NULL); /* time value for retry to make line up */
  fd_idle_last= time(NULL); /* time value for skiping retry to make line up */
  Send_Close_All(i);
  /*
  if(vtun.is_client){
    run_cmd_sh("/etun/r2run", i);
  }
  */
}

int Gap( int t, int base)
{
	int gap;
	gap= t-base;
	if( gap >  MAX_SEQ/2) gap -= MAX_SEQ;
	if( gap < -MAX_SEQ/2) gap += MAX_SEQ;
	return gap;
}

int lfd_linker_udp(void)
{
     register int len, fl, wlen;
     int rseq=0;
     int wseq=0;
     char *obuf[MAX_SEQ+1];
     int owlen[MAX_SEQ+1];
     int rlen[5*2+1];
     fd_set fdset;
     int maxfd, rbuf[5*2+1];
     struct timeval fd_last[5+1], fd_cur, tv, w_last[5+1];
     int i,j;
     int i_max= vtun.svr_num+1;
     int rseq_last=1;
     int next=0;
     int read_count=0;
     int rbuf_first=0;
     unsigned long check_gap=1500;
     int try_lineup[5*2+1];
     send_close_try_count=0;
     fd_idle= time(NULL);
     fd_idle_last= time(NULL)- lfd_host->ka_interval*20;
     lfd_host->ka_failure = 10;
     In=Out=0;
     tm_IO= time(NULL);
     srandom(time(NULL));
     fd[0] = lfd_host->loc_fd; 
     for( i=1; i < i_max; i++) {
	 fd[i]= lfd_host->rmt_fd[i-1];
         gettimeofday(&fd_last[i], NULL);
	 send_close[i]=0;
     }
     for( i=0; i < i_max; i++) {
	idle[i]=-1; /*default to down */
	rbuf[i]=0;
	jitter[i]=0;
	try_lineup[i]=0;
        if( !(buf[i] = lfd_alloc(VTUN_FRAME_SIZE + VTUN_FRAME_OVERHEAD)) ){
          vtun_syslog(LOG_ERR,"Can't allocate buffer for the linker"); 
          return 0; 
        }
     }
     idle[0]=0;
     idle[lfd_host->connected_i]=0;
     live_host_num= 1;
     if( live_host_num == 1) lfd_host->ka_failure= 10; else lfd_host->ka_failure=2;
     for( i=0; i < i_max-1; i++) {if( idle[i+1]== 0) live_host_str[i]= 'U'; else live_host_str[i]= 'D';}
     live_host_str[i]='\0';

     for( i=1; i < i_max; i++) {
       udp_write2(fd[i], buf[i], VTUN_ECHO_REQ, 0);
       gettimeofday(&w_last[i], NULL);
     }

     linker_term = 0;
     while( !linker_term ){
        errno = 0;
        /* Wait for data */
        FD_ZERO(&fdset);
        FD_SET(fd[0], &fdset);
        maxfd = fd[0];
        for( i=1; i < i_max; i++) {
          if(rbuf[i] == 0) FD_SET(fd[i], &fdset);
          maxfd = (fd[i] > maxfd ? fd[i] : maxfd);
        }
        maxfd++;

        tv.tv_sec  = lfd_host->ka_interval;
        tv.tv_usec = 0;

        if( (len = select(maxfd, &fdset, NULL, NULL, &tv)) < 0 ){
           if( errno != EAGAIN && errno != EINTR )
              break;
           else
              continue;
        } 
        /* send ECHO_REP when line is free until check_gap */
        for( i=1; i < i_max; i++) {
           if(idle[i] != -1 ){
             gettimeofday(&fd_cur, NULL);
             timersub(&fd_cur,&w_last[i],&tv);
             if(tv2ms(tv) > check_gap ){
               gettimeofday(&w_last[i], NULL);
               if( udp_write2(fd[i], buf[i], VTUN_ECHO_REP, 0) < 0 ) Idle(i,"echo write error");
             }
           }
        }

        
        /* We are idle, lets check connection (change to just read check) */
        if( lfd_host->flags & VTUN_KEEP_ALIVE  ){
         for( i=1; i < i_max; i++) {
           if( idle[i] != -1 && rbuf[i] == 0 && !FD_ISSET(fd[i], &fdset)){ /* why use rbuf ?? */
             gettimeofday(&fd_cur, NULL);
             timersub(&fd_cur,&fd_last[i],&tv); /* read gap*/
             if(tv2ms(tv) > check_gap ){
              gettimeofday(&fd_last[i], NULL);

              if( ++idle[i] > lfd_host->ka_failure ){
		 Idle(i, "network timeout");
		 /*
                 vtun_syslog(LOG_INFO,"%s Local network timeout line %d, Sending Close All",lfd_host->host, i);
                 idle[i]=0;
                 if(send_close[i] == 0){
	           send_close_try_count++;
	           send_close[i] = 1;
                 } 
                 Send_Close_All(i);
		 if( live_host_num <= send_close_try_count) live_host_num = 0;
                 break;
		 */

              }
             }
           }
         }          
        }

      /* check idle link if server mode*/
      /* send request to client */
      if( !vtun.is_client && ( (time(NULL) - fd_idle) > lfd_host->ka_interval * 10 )){
        for( i=1; i < i_max; i++) {
          if( idle[i] == -1 && try_lineup[i]){
            int k;
            for( k=1; k<i_max; k++){
	     if( idle[k] != -1){
	       sprintf( buf[k], "%d", lfd_host->sopt[k-1].lport);
               vtun_syslog(LOG_INFO,"to_client port %s %d (using %d)", buf[k], i, k);
               if( udp_write2(fd[k], buf[k], VTUN_TO_CLIENT+strlen(buf[k])+1, i) < 0 ) Idle(k,"send request write error");
	       break;
	     }
	    }
	  }
	}
      }

      /* check idle link if client mode*/
      if( vtun.is_client && ( (time(NULL) - fd_idle) > lfd_host->ka_interval * 15 )){
        static int j=0;
	j++;
	if( j> 3000) j=1;
        for( i=1; i < i_max; i++) {
          if( idle[i] == -1){
	   char *str;
	   {
		   str=get_local_addr( lfd_host, i-1);
		   if( strcmp(lfd_host->sopt[i-1].laddr,str) == 0){ /*try with same ip*/
		     if( try_lineup[i]){ /* retry for REQ_TO_SERVER : it will reset when line is up */
                       int k;
                       unsigned port = lfd_host->sopt[i-1].lport;
                       for( k=1; k<i_max; k++){
	                 if( idle[k] != -1){
                           sprintf( buf[k], "%s:%d", str, port); 
                              vtun_syslog(LOG_ERR,"Retry to_server %s %d (use %d)", buf[k], port, i, k);
                           if( udp_write2(fd[k], buf[k], VTUN_TO_SERVER + strlen(buf[k])+1, i) < 0 ) Idle(k,"send request write error");
	                   break;
	                 }
                       }
		     }else{ 
		       /* try run_cmd /etun/run.s1 */
		       if( j%2 == 0){
	                 run_cmd_sh("/etun/r2run", i);
		       }
                       vtun_syslog(LOG_INFO,"Check idle fd[%d]=%d %d", i, fd[i], j);
                       if( udp_write2(fd[i], buf[i], VTUN_ECHO_REQ, 0) < 0 ){
                         vtun_syslog(LOG_INFO,"Check idle fd[%d] error", i);
	               }
		     }
		   }else{
	             try_lineup[i]=1;
                     vtun_syslog(LOG_INFO,"Try New:%s Old:%s", str, lfd_host->sopt[i-1].laddr);
		     free( lfd_host->sopt[i-1].laddr);
		     lfd_host->sopt[i-1].laddr= str;
{ 
     /* first step  (make new socket and port) */
     struct sockaddr_in saddr;
     unsigned port;
     int s,opt;
     if( (s=socket(AF_INET,SOCK_DGRAM,0))== -1 ){
        vtun_syslog(LOG_ERR,"Can't create socket");
        return -1;
     }

     opt=1;
     setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
     saddr.sin_addr.s_addr = inet_addr( lfd_host->sopt[i-1].laddr);
     saddr.sin_port= 0;
     if( bind(s,(struct sockaddr *)&saddr,sizeof(saddr)) ){
        vtun_syslog(LOG_ERR,"Can't bind to the socket");
        return -1;
     }

     opt = sizeof(saddr);
     if( getsockname(s,(struct sockaddr *)&saddr,&opt) ){
         vtun_syslog(LOG_ERR,"Can't get socket name");
         return -1;
     }
     port = saddr.sin_port;
     close(lfd_host->rmt_fd[i-1]); /* close old socket */
     lfd_host->rmt_fd[i-1] = fd[i]= s;
     lfd_host->sopt[i-1].lport = port;

     /* send request to server */
     {
       int k;
       for( k=1; k<i_max; k++){
	 if( idle[k] != -1){
              sprintf( buf[k], "%s:%d", str, port); 
              vtun_syslog(LOG_ERR,"to_server %s %d (use %d)", buf[k], port, i, k);
              if( udp_write2(fd[k], buf[k], VTUN_TO_SERVER + strlen(buf[k])+1, i) < 0 ) Idle(k,"send request write error");
	      break;
	 }
       }
     }
}/* End of Try routine */


		   }
	   } /* end strcmp */
	   
	  }
        }
        fd_idle= time(NULL);
      }

      /* Read frames from network(fd[i]), decode and pass them to 
       * the local device (fd[0]) */
      for( i=1; i <= (i_max-1)*1; i++) {
         if( rbuf[i] == 0 && FD_ISSET(fd[i], &fdset) && lfd_check_up() ){
           if( (len=udp_read2(fd[i], buf[i], &rseq)) <= 0 ) continue;
           gettimeofday(&fd_last[i], NULL);
           fl = len & ~VTUN_FSIZE_MASK;
           len = len & VTUN_FSIZE_MASK;
/*
           vtun_syslog(LOG_INFO,"i:%d, fd[i]:%d, %d idle[i]:%d rseq:%d fl:%x, size:%d", i, fd[i], fd_last[i], idle[i], rseq,fl, len);
           if( fl) vtun_syslog(LOG_INFO,"%s i:%d, fd[i]:%d, idle[i]:%d rseq:%d fl:%x, size:%d", 
			   lfd_host->host, i, fd[i], idle[i], rseq, fl, len);
*/

           /* Handle frame flags */
           if( fl ){
              if( fl==VTUN_ECHO_REP ){
                 /* Just ignore ECHO reply */
                if( idle[i] == -1){
		 if( (time(NULL) - fd_idle_last) > lfd_host->ka_interval * 8 ){ /* del up lack */
		   live_host_num++; 
                   if( live_host_num == 1) lfd_host->ka_failure= 10; else lfd_host->ka_failure=2;
		   live_host_str[i-1]='U';
                   vtun_syslog(LOG_INFO,"%s network up p:%d (%s:%s)", 
				   lfd_host->host, i, lfd_host->sopt[i-1].laddr, lfd_host->sopt[i-1].raddr ); 
		   Set_title();
                   udp_write2(fd[i], buf[i], VTUN_ECHO_REP, 0); 
		   try_lineup[i]=0; /* reset to use again */
		   /*
		   send_close[i]=0;
		   */
                   idle[i] = 0; 
		 }else{
                   vtun_syslog(LOG_INFO,"%s network up SKIP p:%d (%s:%s)", 
				   lfd_host->host, i, lfd_host->sopt[i-1].laddr, lfd_host->sopt[i-1].raddr ); 
		 }
                }else{ 
                 gettimeofday(&fd_last[i], NULL);
                 idle[i] = 0; /* idle 1,2 to 0 with ECHO_REP*/ 
		}
                continue;
              }
              if( fl==VTUN_ECHO_REQ ){
                 /* Send ECHO reply */
                 if( udp_write2(fd[i], buf[i], VTUN_ECHO_REP, 0) < 0 ) Idle(i,"echo_req");
                 continue;
              }
              if( fl==VTUN_STAT_REQ ){
		 /* check qos */
		 static last_live_host_num=0;
	         if( last_live_host_num != live_host_num){
			 run_cmd_sh("/etun/r2qos", live_host_num);
			 last_live_host_num = live_host_num;
		 }
                 /* Send STAT reply */
		 send_log();
                 continue;
	      }
              if( fl==VTUN_STAT_REP ){
		 /*
                 vtun_syslog(LOG_INFO,"from_client stat_rep %s", buf[i] );
		 */
		 log_stat(buf[i]);
                 continue;
	      }
                 /* Set rseq[0]=0 with ECHO SEQ reply for read*/
              if( fl==VTUN_TO_CLIENT ){
                 int i_old=i;
                 int i=rseq;
                 int port;
		 if( !try_lineup[i]) {
                         vtun_syslog(LOG_INFO,"from_server SKIP port %d to make %d (using %d)", port, i, i_old  );
			 continue;
		 }
	       	 port= atoi( buf[i_old]);
                 vtun_syslog(LOG_INFO,"from_server port %d to make %d (using %d)", port, i, i_old  );
		 try_lineup[i]=0;

                 /* try connect */
		 {
                   struct sockaddr_in saddr;
                   int s=lfd_host->rmt_fd[i-1];
                   saddr.sin_addr.s_addr = inet_addr( lfd_host->sopt[i-1].raddr);
                   saddr.sin_port = lfd_host->sopt[i-1].rport= port;

                   if( connect(s,(struct sockaddr *)&saddr,sizeof(saddr)) ){
                     vtun_syslog(LOG_ERR,"Can't connect socket %s:%d", lfd_host->sopt[i-1].raddr, lfd_host->sopt[i-1].rport);
                     return -1;
                   }
                   if( udp_write2(fd[i], buf[i], VTUN_ECHO_REQ, 0) < 0 ){
                     vtun_syslog(LOG_INFO,"Check idle fd[%d] error after connect", i);
	           }
		 }
                 continue;
	      }
              if( fl==VTUN_TO_SERVER ){
                 int i_old=i;
                 int i=rseq;
                 char *str1;
                 int port;
                 str1=strtok(buf[i_old],":");
                 port = atoi(&str1[strlen(str1)+1]);
		 /* if old==new ip then continue; to accept multi try */
		 if( strcmp(lfd_host->sopt[i-1].raddr,str1) == 0){
                         vtun_syslog(LOG_INFO,"from_client SKIP %s:%d to make %d (using %d)", str1, port, i, i_old  );
			 continue;
		 }
                 vtun_syslog(LOG_INFO,"from_client %s:%d to make %d (using %d)", str1, port, i, i_old  );

		 free(lfd_host->sopt[i-1].raddr);
		 lfd_host->sopt[i-1].raddr= strdup(str1);
                 lfd_host->sopt[i-1].rport = port;
		 { /* make s and req to_client */
{ 
     /* first server step  (make new socket and port) */
     struct sockaddr_in saddr;
     unsigned port;
     int s,opt;
     if( (s=socket(AF_INET,SOCK_DGRAM,0))== -1 ){
        vtun_syslog(LOG_ERR,"Can't create socket");
        return -1;
     }

     opt=1;
     setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
     saddr.sin_addr.s_addr = inet_addr( lfd_host->sopt[i-1].laddr);
     saddr.sin_port= 0;
     if( bind(s,(struct sockaddr *)&saddr,sizeof(saddr)) ){
        vtun_syslog(LOG_ERR,"Can't bind to the socket");
        return -1;
     }

     opt = sizeof(saddr);
     if( getsockname(s,(struct sockaddr *)&saddr,&opt) ){
         vtun_syslog(LOG_ERR,"Can't get socket name");
         return -1;
     }
     port = saddr.sin_port;
     close(lfd_host->rmt_fd[i-1]); /* close old socket */
     lfd_host->rmt_fd[i-1] = fd[i]= s;
     lfd_host->sopt[i-1].lport = port;

     /* send request to client */
     {
       int k;
       for( k=1; k<i_max; k++){
	 if( idle[k] != -1){
	      sprintf( buf[k], "%d", port);
              vtun_syslog(LOG_INFO,"to_client port %s %d (using %d)", buf[k], i, k);
              if( udp_write2(fd[k], buf[k], VTUN_TO_CLIENT+strlen(buf[k])+1, i) < 0 ) Idle(k,"send request write error");
	      break;
	 }
       }
     }
     /* try connect */
     saddr.sin_addr.s_addr = inet_addr( lfd_host->sopt[i-1].raddr);
     saddr.sin_port = lfd_host->sopt[i-1].rport;

     if( connect(s,(struct sockaddr *)&saddr,sizeof(saddr)) ){
         vtun_syslog(LOG_ERR,"Can't connect socket %s:%d", lfd_host->sopt[i-1].raddr, lfd_host->sopt[i-1].rport);
         return -1;
     }

}/* End of Try routine */
	         }
                 continue;
	      }
              if( fl==VTUN_BAD_FRAME ){
                 vtun_syslog(LOG_ERR, "%s Received bad frame", lfd_host->host);
                 continue;
              }
              if( fl==VTUN_CONN_CLOSE ){
                char j= rseq; 
                vtun_syslog(LOG_ERR, "%s Received close line %d using %d", lfd_host->host, j, i);
                if( idle[j] != -1) {
		 /*
                 if(send_close_try_count > 0) {
			 send_close_try_count--;
			 send_close[j]=0;
		 }
		 */
		 Idle(j, "close by other side");
                } 
                continue;
              }
           }   
           if( i>=i_max) continue;
           if (idle[i] == -1){
             int t= rseq;
             if( t==0) continue;
	     t= Gap( t, rseq_last); 
	     if( t < 200) {
              if( t> -200) {
                vtun_syslog(LOG_DEBUG, "%s Write Closed fd packet %d, gap:%d", lfd_host->host, rbuf[i], t);
                lfd_host->stat.comp_in += len; 
                if( (len=lfd_run_up(len,buf[i],&out)) == -1 ) Idle(i,"lfd_run_up");
                if( len && dev_write(fd[0],out,len) < 0 ){
                   if( errno != EAGAIN && errno != EINTR ) Idle(i,"local write error");
                   else
                      continue;
                }
                lfd_host->stat.byte_in += len; 
	        lfd_host->stat.packet_count++;
	        In += len;

                rbuf[i]=0;
	      }else{
                 vtun_syslog(LOG_DEBUG, "%s Skip Closed fd packet %d, gap:%d", lfd_host->host, rbuf[i], t);
	         lfd_host->stat.packet_loss ++;
                 rbuf[i]=0;
	      }
	     }
             continue; /* Just ignore when link down status */ 
	   }
           gettimeofday(&fd_last[i], NULL);
           rlen[i]= len;
           rbuf[i]= rseq;
           read_count++;
        }
      }

      /* check for all network is timeout */
      if( live_host_num <= 0){
          vtun_syslog(LOG_INFO,"Session %s network closed all", lfd_host->host);
          break;
      }

      /* skip old packet */
      if( read_count >= live_host_num){ /* read_count is full */
           for( i=1; i < i_max; i++){
            int t= rbuf[i];
            if( t==0) continue;
	    t= Gap( t, rseq_last); 
	    if( t < 0) {
              if( t> -100) {
		/*
                vtun_syslog(LOG_DEBUG, "%s Write Old packet %d, gap:%d", lfd_host->host, rbuf[i], t);
		*/
                len=rlen[i];
                lfd_host->stat.comp_in += len; 
                if( (len=lfd_run_up(len,buf[i],&out)) == -1 ) Idle(i,"lfd_run_up");
                if( len && dev_write(fd[0],out,len) < 0 ){
                   if( errno != EAGAIN && errno != EINTR ) Idle(i,"local write error");
                   else
                      continue;
                }
                lfd_host->stat.byte_in += len; 
	        lfd_host->stat.packet_count++;
	        In += len;

                rbuf[i]=0;
                read_count--;
	      }else{
                 vtun_syslog(LOG_DEBUG, "%s Skip packet %d, gap:%d", lfd_host->host, rbuf[i], t);
	         lfd_host->stat.packet_loss ++;
                 rbuf[i]=0;
                 read_count --;
	      }
	    }
	    
           }
      } /* skip old packet end */

      /* get min rbuf[i] and set to rseq_last if need */
      /* need more hack */
      if( read_count >= live_host_num){
           int r_max=rbuf[1], bad_line=1;
           for( i=2; i< i_max; i++){
              int t_rbuf= rbuf[i];
              if( r_max==0) {r_max=t_rbuf; continue;}
              if( t_rbuf==0) continue;
              if( -MAX_SEQ/2 < t_rbuf-r_max && t_rbuf-r_max < MAX_SEQ/2){
                   if(r_max > t_rbuf) {r_max= t_rbuf; bad_line=i;}
              } else if(r_max < t_rbuf) {r_max= t_rbuf; bad_line=i;}
           }
           
           /* get bad line */
           for( i=1; i<read_count; i++){
              bad_line++;
              if( bad_line >= i_max) bad_line -= i_max-1;
              while( rbuf[bad_line] == 0){
               bad_line++;
               if( bad_line >= i_max) bad_line -= i_max-1;
              }
           }

           if( rseq_last != r_max){
	     lfd_host->stat.packet_loss ++;
	     /*
             if( i_max < 4) vtun_syslog(LOG_DEBUG,"Packet loss seq:%d port:%d ", rseq_last, bad_line);
	     if( i_max == 4)
             vtun_syslog(LOG_DEBUG,"%s loss %d, (%d:%d:%d)%d (%d) gap:%d (%d:%d:%d)", lfd_host->host, rseq_last, rbuf[1], rbuf[2], rbuf[3], bad_line, r_max, Gap(r_max,rseq_last), rlen[1], rlen[2], rlen[3]);
	     if( i_max == 5)
             vtun_syslog(LOG_DEBUG,"%s loss %d, (%d:%d:%d:%d)%d (%d) gap:%d (%d:%d:%d:%d)", lfd_host->host, rseq_last, rbuf[1], rbuf[2], rbuf[3], rbuf[4], bad_line, r_max, Gap(r_max,rseq_last), rlen[1], rlen[2], rlen[3], rlen[4]);
	     */
           }
           rseq_last= r_max;
     }




      /* skip jitter packets */
      if( read_count >= live_host_num){ /* read_count is full */
           for( i=1; i < i_max; i++){
            int t= rbuf[i];
            if( t==0) continue;
	    t= Gap( t, rseq_last); 
	    if ( t > i_max*5){

	     if( i_max == 4)
             vtun_syslog(LOG_DEBUG,"%s jitter %d, (%d:%d:%d)%d gap:%d (%d:%d:%d)", lfd_host->host, rseq_last, rbuf[1], rbuf[2], rbuf[3], i, t, rlen[1], rlen[2], rlen[3]);
	     if( i_max == 5)
             vtun_syslog(LOG_DEBUG,"%s jitter %d, (%d:%d:%d:%d)%d gap:%d (%d:%d:%d:%d)", lfd_host->host, rseq_last, rbuf[1], rbuf[2], rbuf[3], rbuf[4], i, t, rlen[1], rlen[2], rlen[3], rlen[4]);

	         lfd_host->stat.packet_loss ++;
		 rseq_last = rbuf[i];
		 if( ++jitter[i] > 10) {
                         vtun_syslog(LOG_DEBUG,"%s close session by too many jitter line %d count %d", lfd_host->host, i, jitter[i]);
			 live_host_num=0;
		 }
	    }
	   }
      }
      /* End skip jitter packets */
/*
*/
        /* write by packet order when all ready */
        /* get order */
     if( read_count > 0 ){
      for( j=1; j < i_max; j++) {
       for( i=1; i < i_max; i++) {
        if(rbuf[i] == rseq_last){
           len=rlen[i];
           lfd_host->stat.comp_in += len; 
           if( (len=lfd_run_up(len,buf[i],&out)) == -1 ) Idle(i,"lfd_run_up");
           if( len && dev_write(fd[0],out,len) < 0 ){
              if( errno != EAGAIN && errno != EINTR ) Idle(i,"local write error");
              else
                 continue;
           }
           lfd_host->stat.byte_in += len; 
	   lfd_host->stat.packet_count++;
	   In += len;

           rseq_last++;
           if(rseq_last > MAX_SEQ) rseq_last =1;
           rbuf[i]=0;
           read_count--;
        }
        if( read_count <1) break;
       }
       if( read_count <1) break;
      }
     }
     
        /* Read data from the local device(fd[0]), encode and pass it to 
         * the network (fd[i]) */
         if( FD_ISSET(fd[0], &fdset) && lfd_check_down() ){
           /* read */
           if( (wlen = dev_read(fd[0], buf[0], VTUN_FRAME_SIZE)) < 0 ){
              if( errno != EAGAIN && errno != EINTR )
                 idle[0]=-1;
              else
                 continue;
           }
           if( !wlen ) continue;

           /* select next fd (first live fd) */
	   /* full bonding select and skip idle line */
	   {
            do{
                if( ++next > vtun.svr_num) next=1;
	    }while( idle[next]==-1);
	   }


           /* write mode */
           {
            /* lzo, shaper, encrypt etc */
            len=wlen;
            lfd_host->stat.byte_out += wlen; 
	    Out += wlen;
            if((wlen=lfd_run_down(wlen,buf[0],&out)) == -1 ) Idle(i,"lfd_run_down");
            wseq++;
            if( wseq < 1 || wseq > MAX_SEQ) wseq=1;
            if( wlen && udp_write2(fd[next], out, wlen, wseq) < 0 ) Idle(next, "next write error");
            gettimeofday(&w_last[next], NULL); /* to send echo rep */
            lfd_host->stat.comp_out += wlen; 
           }
        }
        Set_IO();
     }
     if( !vtun.is_client ) log_stat(NULL);
     if( !linker_term && errno )
        vtun_syslog(LOG_INFO,"%s (%d)", strerror(errno), errno);

     if (linker_term == VTUN_SIG_TERM) {
       lfd_host->persist = 0;
     }

     /* Notify other end about our close */
     
     for( i=1; i < i_max; i++) {
	     Send_Close_All(i);
     }
     for( i= i_max-1; i >= 0; i--) {
       lfd_free(buf[i]);
     }
     return 0;
}

